#----------------------------------------------------------------
# $Id: //src/racer/Mesh2DOF.py#21 $
# $Date: 2005/06/08 $
#----------------------------------------------------------------
# This file may be copied.  However, if you make improvements,
# please send me a copy.
#
# Brent Burton <brentb@relro.net>
# Originally released: July 2003
# RaceSimCentral: vizor_atx
#----------------------------------------------------------------
#
# Convert a Blender NMesh object to a DOF
# file.  This module needs the DOFfile module.
#

import string
import os
#import Blender
import bpy
import mathutils
#from bpy import Types, Object, NMesh, Material
from bpy import *
#from bpy import bpy.data bpy.ops bpy.props bpy.types bpy.context bgl blf mathutils geometry

from io_export_dof import DOFfile
from io_export_dof import VertOpt

#---------------- module version info
version = 1.0

#----------------
# PRIVATE _DefineMaterial()
# NOTE: formerly (mesh, matidx, ...)
def _DefineMaterial(mat, matidx, matTexNames):
  """
  Return a DOFfile.MAT0 object representing this Blender material.
  """

  # Now convert the mesh material to MAT0 parts...
  print(matidx)
  #mat = mesh.materials[matidx] # now unnecessary

  texfilename = matTexNames[matidx]

  mat0 = DOFfile.MAT0()
  
  mhdr = DOFfile.MHDR()
  mhdr.setName(mat.name)
  mat0.setMHDR(mhdr)

  mcol = DOFfile.MCOL()
  #mcol.setAmbient([mat.R*mat.amb, mat.G*mat.amb, mat.B*mat.amb, mat.alpha])
  mcol.setAmbient([mat.ambient, mat.ambient, mat.ambient, mat.alpha])
  #mcol.setDiffuse([mat.R, mat.G, mat.B, mat.alpha])
  mcol.setDiffuse([mat.diffuse_color[0], mat.diffuse_color[1], mat.diffuse_color[2], mat.alpha])
  #mcol.setSpecular([mat.specCol[0]*mat.spec, mat.specCol[1]*mat.spec, mat.specCol[2]*mat.spec, mat.alpha])
  mcol.setSpecular([mat.specular_color[0]*mat.specular_intensity, mat.specular_color[1]*mat.specular_intensity, mat.specular_color[2]*mat.specular_intensity, mat.alpha])
  #mcol.setEmission([mat.R*mat.emit, mat.G*mat.emit, mat.B*mat.emit, mat.alpha])
  mcol.setEmission([mat.emit, mat.emit, mat.emit, mat.alpha])
  #mcol.setShininess(mat.hard)
  mcol.setShininess(mat.specular_hardness)
  mat0.setMCOL(mcol)

  mcfl = DOFfile.MCFL()
  mat0.setMCFL(mcfl)

  muvw = DOFfile.MUVW()
  muvw.setUVOffset([0.0, 0.0])
  muvw.setUVTiling([1.0, 1.0])
  muvw.setAngle(0.0)
  muvw.setBlur(1.0)
  muvw.setBlurOffset(0.0)
  mat0.setMUVW(muvw)

  mtra = DOFfile.MTRA()
  mtra.setTransparency(1 - mat.alpha) # unused field, used for BLEND_CONST I guess.
  if mat.alpha > 0.99:
    mtra.setBlendmode(0)            # off
  else:
    mtra.setBlendmode(1)            # 1=BLEND_SRC_ALPHA; 2=BLEND_CONST(unused)
  mat0.setMTRA(mtra)

  mtex = DOFfile.MTEX()
  mtex.append(texfilename)
  mat0.setMTEX(mtex)
  return mat0


#----------------
# PRIVATE _DefineBogusMaterial()
def _DefineBogusMaterial(matidx, matTexNames):
  """
  Return a DOFfile.MAT0 object for a texture's name.
  """

  texfilename = matTexNames[matidx]
  # Make material name by stripping off texture filename extension
  #radome matname = texfilename.join(string.split(texfilename, ".")[0:-1], "")
  matname = texfilename.partition('.')[0]
  #print(matname)
  
  mat0 = DOFfile.MAT0()

  mhdr = DOFfile.MHDR()
  mhdr.setName(matname)
  mat0.setMHDR(mhdr)

  mcol = DOFfile.MCOL()
  mcol.setAmbient([0.5, 0.5, 0.5, 1.0])
  mcol.setDiffuse([0.8, 0.8, 0.8, 1.0])
  mcol.setSpecular([0.0, 0.0, 0.0, 1.0])
  mcol.setEmission([0.0, 0.0, 0.0, 1.0])
  mcol.setShininess(128.0)
  mat0.setMCOL(mcol)

  mcfl = DOFfile.MCFL()
  mat0.setMCFL(mcfl)

  muvw = DOFfile.MUVW()
  muvw.setUVOffset([0.0, 0.0])
  muvw.setUVTiling([1.0, 1.0])
  muvw.setAngle(0.0)
  muvw.setBlur(1.0)
  muvw.setBlurOffset(0.0)
  mat0.setMUVW(muvw)

  mtra = DOFfile.MTRA()
  mtra.setTransparency(0.0)           # unused
  mtra.setBlendmode(0)                # off
  mat0.setMTRA(mtra)

  mtex = DOFfile.MTEX()
  mtex.append(texfilename)
  mat0.setMTEX(mtex)
  return mat0


# Object coordinates are local.  Each object's reference
# point can be moved in Blender to translate the entire object.
# Further, scaling can take place.  Sooo, adjust XYZ
# coords by appropriate scalings and translations.
# The scale arg is a final inverse-modeling scale.
# TODO: do matrix rotations
# Stereo says: swapped out the original scale/location for the matrix_world
# bonus: rotation happens for free.
def _xformVertex(v, bobj, scale=1.0):
  #x = (v[0] * bobj.scale[0] + bobj.location[0]) * scale
  #y = (v[1] * bobj.scale[1] + bobj.location[1]) * scale
  #z = (v[2] * bobj.scale[2] + bobj.location[2]) * scale
  vout = bobj.matrix_world*v
  # the (x, z, -y) is intentional as it converts
  # the modeler (3dsmax/blender) coordinate system
  # to OpenGL's.
  #return [x, z, -y]
  return [vout[0], vout[2], -vout[1]]

def _xformNormal(n, bobj):
  # Ditto on the x,z,-y thing here
  nout = bobj.matrix_world*n
  nout.normalize()
  return [nout[0], nout[2], -nout[1]]

def _getVColor(nmcol):
  # NMCol input
  r = nmcol[0] / 255.0
  g = nmcol[1] / 255.0
  b = nmcol[2] / 255.0
  return [r, g, b]

def _insertVertex(vindex, obj, face, uvface, verts, vo, gob, scale, exportVertColors=False):
  xyz    = (x,y,z)    = verts[face.vertices_raw[vindex]].co
  uv     = (u,v)      = uvface.uv[vindex]
  normal = (xn,yn,zn) = verts[face.vertices_raw[vindex]].normal
  idx,newvert = vo.addVertex((x,y,z,u,v,xn,yn,zn))
  if newvert:
    gob._vert.append(_xformVertex(xyz, obj, scale))
    gob._norm.append(_xformNormal(normal,obj))
    if exportVertColors: 
      gob._vcol.append(_getVColor(obj.color))
    gob._tver.append(uv)
  return idx

#----------------
# PRIVATE ConvertMeshToDOF():
def _ConvertMeshToDOF(obj, scale=1.0, use_mesh_modifiers=False):
  # radome
  # Stereo: todo - switch obj.data for the modifier friendly
  # which creates a temp object, drops it in an array, then, later...
  # if use_mesh_modifiers:
  #   mesh = obj.to_mesh(scene,True,'PREVIEW')
  #   meshes_to_clear.append(mesh)
  # [...]
  #  for mesh in meshes_to_clear:
  #    bpy.data.meshes.remove(mesh)
  print(obj.data)
  mesh = obj.data
  # Create the DOF.  Then add a Material
  # and the geometry
  dof1 = DOFfile.DOF1()
  # If this is set to 1 AND a mesh doesn't have a material
  # defined, vertex colors are written out to the DOF to
  # provide coloring.  This disables Racer's lighting
  # calculations on the meshes that have vertex coloring.
  # However, if it is forced to 0 OR the mesh has materials
  # defined, no vertex colors are exported and Racer shades the
  # mesh with lighting, based on the material's properties such
  # as reflectance, specular color, ambient color and emissive
  # color.
  exportVertColors = 0

  # Preprocess materials to size lists
  numMats = len(mesh.materials)
  if numMats > 0:
    exportVertColors = 0
    matFaces = []                       # array of empty lists
    matTexNames = []
    for i in range(numMats):
      matFaces.append([])             # add another empty list
      matTexNames.append("image_missing")        # add placeholder
    
    # Now figure out which faces go with which materials.
    # Changed for Blender 2.63 and up -- Stereo.
    # mesh.faces is gone, mesh.tessfaces needs to be prepared
    mesh.calc_tessface()
    lastMatIndex = -1
    for i in range(len(mesh.tessfaces)):
      print(mesh.tessfaces[i])
      matidx = mesh.tessfaces[i].material_index
      print(matidx)
      matFaces[matidx].append(i)
      if lastMatIndex != matidx:
        #if mesh.tessfaces[i].image is None:
        if mesh.materials[matidx].active_texture is None:
          print ('*** Face %d has no texture!!' % (i))
          continue
        #matTexNames[matidx] = string.lower(mesh.tessfaces[i].image.getName())
        matTexNames[matidx] = mesh.materials[matidx].active_texture.image.filepath_raw.lstrip('//') or "none"
        lastMatIndex = matidx

    # Finally define the actual mats since texnames are now known
    for i in range(numMats):
      mat0 = _DefineMaterial(mesh, i, matTexNames)
      dof1.addMAT0(mat0)

  else:
    # radome sez - This area needs a bit of work.  I worked on it back in June.
    # The program goes here if you have --
    
    # No materials defined.  The intent is to use Racer shaders
    # so iterate through the faces and see which textures have
    # been applied.  At least that much should be defined...
    lastMatIndex = -1
    texNamesIndex = {}
    matFaces = []                   # array of empty lists
    matTexNames = []
    print(mesh.active_uv_texture.data[0].image.name)
    for i in range(len(mesh.active_uv_texture.data)):
      # radome if mesh.tessfaces[i].image is None: 
      if mesh.active_uv_texture.data[i].image is None:
        print ('*** Face %d has no texture!!' %(i))
        continue
      texname = mesh.active_uv_texture.data[i].image.name
      # Map texture name to an index; define one if not mapped.
      if texname not in texNamesIndex:
        # radome added list() based on 2to3.py recommendation
        if len(list(texNamesIndex.keys())) == 0: # if nothing in it yet
          idx = 0     # start index at zero
        else:
          idx = max(texNamesIndex.values()) + 1
        matFaces.append([])
        matTexNames.append(None)
        texNamesIndex[texname] = idx

      matidx = texNamesIndex[texname]
      matFaces[matidx].append(i)  # append face number to this material's list
      if lastMatIndex != matidx:
        matTexNames[matidx] = texname
        lastMatIndex = matidx

    # Define bogus materials for these texture names
    for i in range(len(matTexNames)):
      mat0 = _DefineBogusMaterial(i, matTexNames)
      dof1.addMAT0(mat0)

  #-- Materials are now defined.

  # Init some counters, and write the geometry out:
  #verts = mesh.verts
  verts = mesh.vertices
  numOriginalVerts = len(verts)
  totalIndexCount = 0
  totalNumTris = 0
  totalOptIndexCount = 0

  # For each material, make a new geob and process its faces
  for i in range(len(matFaces)):
    #print("mat " + i + " has " + len(matFaces[i]) + " faces.")
    if len(matFaces[i]) == 0:       # no faces for material i
      continue

    gob1 = DOFfile.GOB1()
    dof1.addGOB1(gob1)

    gob1.setGHDR( DOFfile.GHDR(0,0,i) )

    # Iterate over the faces.  Vertices have XYZ and UV data
    # separate in Blender; faces pick XYZ and UV values via
    # indices, and in Racer the XYZ/UV needs to be combined.
    # Since a single XYZ value can have numerous UV values
    # (in the case of corner (i.e. shared) vertices), the XYZ
    # values must be duplicated for as many faces that share them...
    dofverts   = DOFfile.VERT()
    dofvcols   = DOFfile.VCOL()
    dofuvs     = DOFfile.TVER()
    dofnorms   = DOFfile.NORM()
    dofindices = DOFfile.INDI()
    dofburst   = DOFfile.BRST()

    origIndexCount = 0
    optIndexCount = -1
    numTris = 0
    vo = VertOpt.VO()               # create Vertex Optimizer
    for faceIndex in matFaces[i]:
      #face = mesh.tessfaces[faceIndex]
      face = mesh.tessfaces[faceIndex]
      textureindex = 0
      # another change in 2.63: tessface_uv_textures -- Stereo
      if len(mesh.tessface_uv_textures) > 1:
        while not mesh.tessface_uv_textures[textureindex].active:
          textureindex += 1
      uvface = mesh.tessface_uv_textures[textureindex].data[faceIndex]
      if face.vertices_raw[2] != 0: 
        corners = 3
        if face.vertices_raw[3] != 0:
          corners = 4
      #corners = len(face.raw_verts)
      if corners >= 3:                # add 1st tri: 0,1,2
                
        # XYZ, Norm, UV
        
        xyz    = (x,y,z)    = verts[face.vertices_raw[0]].co
        uv     = (u,v)      = uvface.uv1
        normal = (xn,yn,zn) = verts[face.vertices_raw[0]].normal
        idx0 = vo.addVertex((x,y,z,u,v,xn,yn,zn))
        if idx0 > optIndexCount:
          optIndexCount = idx0
          dofverts.append(_xformVertex(xyz, obj, scale))
          dofnorms.append(_xformNormal(normal))
          dofvcols.append(_getVColor(obj.color))
          dofuvs.append(uv)

        xyz    = (x,y,z)    = verts[face.vertices_raw[1]].co
        normal = (xn,yn,zn) = verts[face.vertices_raw[1]].normal
        uv     = (u,v)      = uvface.uv2
        idx1 = vo.addVertex((x,y,z,u,v,xn,yn,zn))
        if idx1 > optIndexCount:
          optIndexCount = idx1
          dofverts.append(_xformVertex(xyz, obj, scale))
          dofnorms.append(_xformNormal(normal))
          dofvcols.append(_getVColor(obj.color))
          dofuvs.append(uv)

        xyz    = (x,y,z)    = verts[face.vertices_raw[2]].co
        normal = (xn,yn,zn) = verts[face.vertices_raw[2]].normal
        uv     = (u,v)      = uvface.uv3
        idx2 = vo.addVertex((x,y,z,u,v,xn,yn,zn))
        if idx2 > optIndexCount:
          optIndexCount = idx2
          dofverts.append(_xformVertex(xyz, obj, scale))
          dofnorms.append(_xformNormal(normal))
          dofvcols.append(_getVColor(obj.color))
          dofuvs.append(uv)

        origIndexCount += 3
        dofindices.append(idx0)
        dofindices.append(idx1)
        dofindices.append(idx2)
        numTris += 1

      if corners == 4:                # add 2nd tri: 2,3,0
        xyz    = (x,y,z)    = verts[face.vertices_raw[3]].co
        normal = (xn,yn,zn) = verts[face.vertices_raw[3]].normal
        uv     = (u,v)      = uvface.uv4
        idx3 = vo.addVertex((x,y,z,u,v,xn,yn,zn))
        if idx3 > optIndexCount:
          optIndexCount = idx3
          dofverts.append(_xformVertex(xyz, obj, scale))
          dofnorms.append(_xformNormal(normal))
          dofvcols.append(_getVColor(obj.color))
          dofuvs.append(uv)

        origIndexCount += 1
        dofindices.append(idx2)     # reuse
        dofindices.append(idx3)
        dofindices.append(idx0)     # reuse
        numTris += 1

    del vo

    dofburst.addBurst(0, numTris * 9, 0, 3)
    totalNumTris += numTris
    totalIndexCount += origIndexCount
    totalOptIndexCount += optIndexCount

    # Done.  Pass the arrays to the geom object.
    gob1.setVERT(dofverts)
    if exportVertColors:  gob1.setVCOL(dofvcols)
    gob1.setNORM(dofnorms)
    gob1.setTVER(dofuvs)
    gob1.setINDI(dofindices)
    gob1.setBRST(dofburst)
    # End of faces for this material

  return dof1, numOriginalVerts, totalIndexCount, totalOptIndexCount, totalNumTris


#----------------
# PUBLIC ExportMesh()
def ExportMesh(obj, directory, scale=1.0):
  """Exports a Blender NMesh object to a DOF file in 'directory'."""

  (dof1, vin, vout, vopt, tris) = _ConvertMeshToDOF(obj, scale)
  dirspec = directory.replace("\\", "/")
  lastChar = dirspec[len(dirspec)-1]
  if lastChar != "/": dirspec += "/"

  #filespec = (dirspec + filename.replace(obj.name, ".", "") + ".dof").lower
  filespec = dirspec + obj.name + ".dof"
  filespec = filespec.lower()
  outf = open(filespec, "wb")
  dof1.write(outf)
  outf.flush()
  outf.close()
  return (vin, vout, vopt, tris)

#----------------
# PUBLIC ExportLayer()
# Scale is an inverse-modeling factor.
# I.e., if the modeling in Blender is done at 1:10, scale would be 10.0.
def ExportLayer(layer, directory, scale=1.0):
  """Exports all mesh objects from a particular layer.
  If layer==-1, all mesh objects are exported.
  The resulting DOF files are stored in 'directory'.
  """
  numExported = 0
  totalVin = 0
  totalVout = 0
  totalVopt = 0
  totalTris = 0
  for obj in bpy.data.objects:
    print(obj)
    if layer == -1 or obj.layers[layer - 1] == True:
      if obj.type == "MESH":
        print ("Exporting mesh " +  obj.name)
        (vin, vout, vopt, tris) = ExportMesh(obj, directory, scale)
        print ("---> Vin: %d  Vout: %d  Vopt: %d Tris: %d" % (vin, vout, vopt, tris))
        totalVin += vin
        totalVout += vout
        totalVopt += vopt
        totalTris += tris
        numExported += 1
        print ("%d meshes exported into %s" % (numExported, directory))
        print ("Totals: Vin: %d  Vout: %d  Vopt: %d Tris: %d" % (totalVin, totalVout, totalVopt, totalTris))

# -------------------
# PUBLIC export_dof()
# This one's just triage I suppose.  Decides what objects to put in each dof.
def export_dof(operator, context,
               filepath="",
               use_selection=False,
               batch_mode=False,
               **kwargs
               ):
  
  if bpy.ops.object.mode_set.poll():
    bpy.ops.object.mode_set(mode='OBJECT')
    
  if not batch_mode:
    kwargs_mod = kwargs.copy()
    if use_selection:
      kwargs_mod['context_objects'] = context.selected_objects
    else:
      kwargs_mod['context_objects'] = context.scene.objects
    return save_single(operator, context.scene, filepath, **kwargs_mod)

# --------------------
# PUBLIC save_single()
# Saves the given object list into one file, which needs to be an argument.
def save_single(operator, scene, filepath='',
  use_mesh_modifiers=True,
  parent_translate=False,
  plus_ninety=False,
  context_objects=None,
  exportVertColors=False,
  scale=1.0,
  **kwargs
  ):
  # I don't think I need this actually.
  #base_dst = os.path.dirname(filepath)
  print('\nDOF export starting, into: %r' % filepath)
  # Create the DOF.  Then add a Material
  # and the geometry
  dof1 = DOFfile.DOF1()
  GOBs = []                # array of GOBs to write out
  gobVO = []               # array of vertex optimizers for each GOB
  nGOBs = 0
  matGOBs = []             # array of GOB indices to write to for each material (in case mat has 2+)
  matNames = []             # array of material names (for lookup) - blanked when GOB is full
  matTexNames = []         # array of textures used on materials
  
  # Stereo: todo - switch obj.data for the modifier friendly
  # which creates a temp object, drops it in an array, then, later...
  # if use_mesh_modifiers:
  #   mesh = obj.to_mesh(scene,True,'PREVIEW')
  #   meshes_to_clear.append(mesh)
  # [...]
  #  for mesh in meshes_to_clear:
  #    bpy.data.meshes.remove(mesh)
  meshes_to_clear = []
  for obj in context_objects:
    print(obj.name)
    # try to coerce object into a mesh.
    try:
      mesh = obj.to_mesh(scene,use_mesh_modifiers,'PREVIEW')
      if parent_translate:
        par = obj.parent
        if par:
          ploc = par.location
          mesh.transform(mathutils.Matrix.Translation(-ploc))
          mesh.update()
          mesh.transform(mathutils.Euler(par.rotation_euler).to_matrix().to_4x4().transposed())
          mesh.update()
          if plus_ninety:
            mesh.transform(mathutils.Matrix.Rotation(1.5707963267948966,4,'X'))
            mesh.update()
      meshes_to_clear.append(mesh)
    except:
      continue
    verts = mesh.vertices
    # I've removed the vert colors code for now -- Stereo
    # Preprocess materials to size lists
    numMats = len(mesh.materials)
    if numMats > 0:
      matMaps = []
      for i in range(numMats):
        try:
          matMaps.append(matNames.index(mesh.materials[i].name))
        except:
          matidx = len(matNames) # not matNames.index(mesh.materials[i].name) - that's circular dependent
          matGOBs.append(len(GOBs))         # set the material's GOB to the next available index
          gob1 = DOFfile.GOB1()             # add another empty GOB
          gob1.setGHDR( DOFfile.GHDR(0,0,matidx) ) # set its material header up
          gob1.setVERT(DOFfile.VERT())                         # gob._vert
          if exportVertColors:  gob1.setVCOL(DOFfile.VCOL())   # gob._vcol
          gob1.setNORM(DOFfile.NORM())                         # gob._norm
          gob1.setTVER(DOFfile.TVER()) # uvs                     gob._tver
          gob1.setINDI(DOFfile.INDI())                         # gob._indi
          gob1.setBRST(DOFfile.BRST())                         # gob._brst
          GOBs.append(gob1)
          gobVO.append(VertOpt.VO())               # create Vertex Optimizer
          matMaps.append(matidx)
          matNames.append(mesh.materials[i].name)
          matTexNames.append("%s.tga" % mesh.materials[i].name)        # add placeholder

      # Now add each material's faces into the correct GOB1.
      mesh.calc_tessface() # not strictly necessary but sometimes good
      lastMatIndex = -1
      # TODO: possibility of adding support for more UV maps goes here.  For now, only use active one.
      # TODO: support for objects with no UV map at all (maybe an array with zeros?)
      for face, uvface in zip(mesh.tessfaces, mesh.tessface_uv_textures.active.data):
        #print(face)
        omatidx = face.material_index # mesh's mat index
        #print(omatidx)
        matidx = matMaps[omatidx] # output file's mat index
        gobidx = matGOBs[matidx]  # output file's gob index
        #matFaces[matidx].append(i)
        # TODO: add check for GEOBs with more than 20k vertices, in which case spawn new gob1.
        corners = 0
        if face.vertices_raw[2] != 0: 
          corners = 3
          if face.vertices_raw[3] != 0:
            corners = 4
        if corners >= 3:
          # TODO: instead of (face,verts) just pass verts[face[i]]? not sure if better
          idx0 = _insertVertex(0, obj, face, uvface, verts, gobVO[gobidx], GOBs[gobidx], scale)
          idx1 = _insertVertex(1, obj, face, uvface, verts, gobVO[gobidx], GOBs[gobidx], scale)
          idx2 = _insertVertex(2, obj, face, uvface, verts, gobVO[gobidx], GOBs[gobidx], scale)
          GOBs[gobidx]._indi.append(idx0)
          GOBs[gobidx]._indi.append(idx1)
          GOBs[gobidx]._indi.append(idx2)
          if corners == 4:
            idx3 = _insertVertex(3, obj, face, uvface, verts, gobVO[gobidx], GOBs[gobidx], scale)
            # the second triangle shares 2 vertices so no need to recalculate.
            GOBs[gobidx]._indi.append(idx2)
            GOBs[gobidx]._indi.append(idx3)
            GOBs[gobidx]._indi.append(idx0)
        ''' not sure what this code is for.  pulling texture names for each material I guess.
        if lastMatIndex != matidx:
          #if mesh.tessfaces[i].image is None:
          if mesh.materials[matidx].active_texture is None:
            print ('*** Face %d has no texture!!' % (i))
            continue
          #matTexNames[matidx] = string.lower(mesh.tessfaces[i].image.getName())
          matTexNames[matidx] = mesh.materials[matidx].active_texture.image.filepath_raw.lstrip('//') or "none"
          lastMatIndex = matidx'''

      '''# Finally define the actual mats since texnames are now known
      for i in range(numMats):
        mat0 = _DefineMaterial(mesh, i, matTexNames)
        dof1.addMAT0(mat0)'''

    else:
      # Every object must have a material.
      operator.report({'ERROR'}, "Object %r requires at least one material." % obj.name)
  
  # this is after looping through all objects producing geobs & mats.
  for idx,mat in enumerate(matNames):
    mat0 = _DefineMaterial(bpy.data.materials[mat], idx, matTexNames)
    dof1.addMAT0(mat0)
  for gob in GOBs:
    # the docs say brst's 2nd value should be 3*number vertices, instead
    # the exporter is using 9*number triangles, ie. 3*number indices
    # docs also say the 3rd value should be the mtl idx of this geob
    # but instead this is just using 0.  so idk.
    print("GOB1 with {} triangles, {} vertices.".format(len(gob._indi._indices)/3, len(gob._vert._verts)))
    gob._brst.addBurst(0, len(gob._indi._indices) * 3, 0, 3)
    dof1.addGOB1(gob)
  for vo in gobVO:
    del vo
  #-- dof is now ready.

  try:
    outf = open(filepath, "wb")
    dof1.write(outf)
    outf.flush()
    outf.close()
  except:
    import traceback
    traceback.print_exc()
    operator.report({'ERROR'}, "Couldn't write to file %r" % filepath)
    return {'CANCELLED'}
  
  for mesh in meshes_to_clear:
    bpy.data.meshes.remove(mesh)  
  return {'FINISHED'} # does something or other

# EOF