# This integrates mesh2dof into the Blender interface.

bl_info = {
  "name":"Racer DOF format",
  "author": "Brent Burton, N Murdoch",
  "location": "File > Export",
  "description": "Export DOF for Racer.  Meshes, UVs, materials",
  "version": (0,1),
  "blender": (2.63,0),
  "support":"COMMUNITY",
  "category":"Import-Export"
  }

if "bpy" in locals():
  import imp
  imp.reload(DOFfile)
  imp.reload(Mesh2DOF)
  imp.reload(VertOpt)
  print("Reloaded dof export.")
else:
  from . import DOFfile, Mesh2DOF, VertOpt
  print("Imported dof export.")

import bpy
from bpy.props import (StringProperty,
                       BoolProperty,
                       FloatProperty,
                       )
from bpy_extras.io_utils import (ExportHelper,
                                 path_reference_mode,
                                 )

class ExportDOF(bpy.types.Operator, ExportHelper):
  """Export a DOF geometry file"""
  bl_idname="export_scene.dof"
  bl_label="Export DOF"
  # bl_options =   # not sure what these do, will skip
  filename_ext = ".dof"
  filter_glob = StringProperty(default='*.dof',options={'HIDDEN'})
  # list of operator properties, Blender does magic to these to
  # get the UI later on out of an instance of the class
  use_selection = BoolProperty(
    name="Selected Objects",
    description="Export selected objects on visible layers",
    default=False,
  )
  use_mesh_modifiers = BoolProperty(
    name="Apply Modifiers",
    description="Apply modifiers to mesh objects",
    default=True,
  )
  parent_translate = BoolProperty(
    name="Use Parent Origin",
    description="Subtracts parent's coordinates from object to translate geometry to the correct location",
    default=False,
  )
  plus_ninety = BoolProperty(
    name="Plus 90 Around X",
    description="If parent origin is active, rotates objects 90 degrees - compensates for model's preferred 'rotate around z' where Racer rotates around y",
    default=False,
  )
  batch_mode = BoolProperty(
    name="Batch Mode",
    description="Exports a separate file for each root level empty available",
    default=False,
    options={"HIDDEN"},
  )
  def draw(self, context):
    layout = self.layout
    layout.prop(self,"use_selection")
    layout.prop(self,"use_mesh_modifiers")
    layout.prop(self,"parent_translate")
    layout.prop(self,"plus_ninety")
  @property
  def check_extension(self):
    return not self.batch_mode
  def execute(self, context):
    if not self.filepath:
      raise Exception("filepath not set")
    from . import Mesh2DOF
    keywords = self.as_keywords()
    return Mesh2DOF.export_dof(self, context, **keywords)

def menu_func_export(self,context):
  self.layout.operator(ExportDOF.bl_idname, text="Racer DOF (.dof)")

def register():
  bpy.utils.register_module(__name__)
  bpy.types.INFO_MT_file_export.append(menu_func_export)

def unregister():
  bpy.utils.unregister_module(__name__)
  bpy.types.INFO_MT_file_export.remove(menu_func_export)

if __name__ == "__main__":
  register()